/*
 * Goldwater Distributed Web Architecture.
 *
 * Admin Command Processing
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:10 $
 * 
 * --------------------------------------------------------------------------
 * $Log: command.c,v $
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.19  2003/03/26 19:22:35  chris
 * Added cycle command.
 *
 * Revision 1.18  2002/12/01 13:08:08  chris
 * Removed force into interactive mode and added options to perform a
 * pre-command before entering batch mode, interactive mode or command mode
 * (yes you can execute two command with goldwater -c 'command1' command2)
 *
 * Revision 1.17  2002/11/30 20:02:46  chris
 * blank line after displaying set action
 *
 * Revision 1.16  2002/11/30 20:01:26  chris
 * Prompt verbose was wrong.,
 *
 * Revision 1.15  2002/11/30 20:00:01  chris
 * Work to allow forcing into interactive mode.
 *
 * Revision 1.14  2002/11/30 19:40:53  chris
 * Fixed a couple of formatting bugs wrt prompt changing.
 *
 * Revision 1.13  2002/11/30 19:34:38  chris
 * Allow setting of administration prompt.
 *
 * Revision 1.12  2002/11/30 14:04:42  chris
 * GPL headers
 *
 * Revision 1.11  2002/10/20 22:45:47  chris
 * Cleaned set loglevel display
 *
 * Revision 1.10  2002/10/20 13:38:30  chris
 * Enabled setting of loglevel from administrator (removed _LogLevel and
 * replaced it with *_LogLevelPtr through out code).
 *
 * Revision 1.9  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.8  2002/05/21 21:15:16  chris
 * Big tidy up - lots of misc changes.
 * Changed 'internal' services to be prefixed by '!' so they don't show in
 * the admin tool service list.
 *
 * Revision 1.7  2002/03/16 17:28:49  chris
 * Queue listing.
 *
 * Revision 1.6  2002/03/15 19:36:03  chris
 * Display 'command not implemented' where appropriate.
 * Adjust total number of services to not report those which are private.
 * New version number.
 *
 * Revision 1.5  2002/03/15 19:28:48  chris
 * Added Service List and 'pager' support via more/less ($PAGER)
 *
 * Revision 1.4  2002/03/14 21:59:09  chris
 * Queue stats command
 *
 * Revision 1.3  2002/03/14 21:49:23  chris
 * Process listing
 *
 * Revision 1.2  2002/03/10 15:15:32  chris
 * Admin tool changes
 *
 * Revision 1.1  2002/02/26 00:30:20  chris
 * All new goldwater admin tool source.
 *
 * ---------------------------------------------------------------------------
 */

#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <phlib/xl_types.h>

#include <gwlib/gwgbprm.h>
#include "command.h"

extern int exit_flag;
extern char admin_prompt[];

RetValue log_string( CString * strCS ) { printf( "Command not implemented %.*s\n\n", CSLenPtr(strCS), CSStartPtr(strCS) ); return Success;}

/* ------------------------------------------------------------------------- */

RetValue 
gw_help( CString * argCS ) 
{
  int i = 0;

  printf( "\nGoldwater commands:\n" );

  while( cmndtab[i].t_text != NULL ) {
    printf( "%-11.118s %s\n", cmndtab[i].t_text, cmndtab[i].doc );
    i++;
  }
  printf("\n");

  return Success;
}

/* ------------------------------------------------------------------------- */

RetValue 
gw_stats_list( CString * argCS )
{
  return log_string( argCS );
}

/* ------------------------------------------------------------------------- */

RetValue 
gw_quit_admin( CString * argCS )
{
  exit_flag = 1;
  return Success;
}

/* ------------------------------------------------------------------------- */

RetValue 
gw_set_command( CString * argCS )
{
  int i = 0;
  long along = 0;

  CString varCS = CSInit;
  CString longCS = CSInit;
  CString strCS  = CSInit;
  VariableList *ventry = NULL;

  skip_lead_space( argCS );
  skip_tail_space( argCS );

  chop_word( &varCS, argCS );

  skip_tail_space( &varCS );

  if( CSLen(varCS) == 0 ) {
    printf( "Syntax error: <variable> missing from expression\n\n" );
	return Failure;
  }
  if( CSLenPtr(argCS) == 0 ) {
    printf( "Syntax error: <value> missing from expression\n\n" );
	return Failure;
  }

  skip_lead_space( argCS );

  /* Scan variable table for variable
   */
  while( vartable[i].t_text != NULL ) {
    if( strcmpCS( &varCS, vartable[i].t_text) == 0 ) {
	  ventry = &vartable[i]; 
	  break;
	}
    i++;
  }

  if( !ventry ) {
    printf( "Unknown variable [%.*s]\n\n", CSLenPtr(&varCS), CSStartPtr(&varCS));
    return Success;
  }


  /* Consider type of variable
   */
  switch( ventry->type )
  {
    case VAR_LONG:
      chop_word( &longCS, argCS );
	  if( CSLenPtr(argCS) > 0 ) {
	    printf( "Extraneous characters in expression\n\n" );
		return Failure;
	  }
	  if( CSLen(longCS) == 0 ) {
	    printf( "Missing value in expression\n\n" );
		return Failure;
	  }
	  if( CStoInt32( &longCS, &along ) == Failure ) {
	    printf( "Invalid value in expression\n\n" );
		return Failure;
	  }
	break;

	case VAR_STRING:
	  MirrorCStringPtr( &strCS, argCS );
	break;
  }

  switch( ventry->var ) {
  case V_LOGLEVEL:
    *_LogLevelPtr = along;
    printf( "%s set to %ld\n", ventry->doc, along );
	break;

  case V_REQTOUT:
    printf( "%s set to %ld\n", ventry->doc, along );
    break;

  case V_PROMPT:
    CStoCharCpy( admin_prompt, &strCS, PROMPT_MAX );
	/* Don't need this as the change will be obvious! */
    /* printf( "%s set to '%s'\n", ventry->doc, admin_prompt ); */
  }
  printf( "\n" );

  return Success;
}

/* ------------------------------------------------------------------------- */
/* end */

