/*
 * Goldwater Distributed Web Architecture.
 * - Goldwater Cycle Servers Command
 *
 * Copyright 1996-2003 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.1 $  $Date: 2003/03/29 12:21:10 $
 * 
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_linklst.h>

#include <gwreq.h>
#include <gwlog.h>
#include <gwclient.h>
#include <gwapi.h>

/* Privates */
#include "goldreq.h"
#include "regstrings.h"
#include "gwreadconf.h"
#include "gwgbprm.h"
#include "mutex.h"

#include "gwadmin.h"

#define INV_CLASS  "ERR: Invalid Class Identifier\n"
#define INV_SERVER "ERR: Invalid Server Identifier\n"
#define CYCLE_USAGE "Usage: cycle -s server|-c class id -i service id|-c class id\
n\n"

static RetValue cycle_server( long class, long instance );
static RetValue compile_cycle_list( long classid, long srvid, LIST *hlist );

RetValue kill_server( BootParams *bp );

RetValue launch_server( BootParams *bp );

extern char *local_dom;
extern char  local_dom_lev[];

/* ------------------------------------------------------------------------- */

RetValue
gw_cycle( CString *argCS )
{
  long  val       = 0;
  CString optvCS = CSInit;
  char    opt    = '0';
  int     go     = 1;
  long    class  = 0;
  long    id     = 0;
  char    server[128] = {0};


  while( CSLenPtr( argCS ) > 0 ) {
    
    if( parse_arguments( argCS, &optvCS, &opt, "s:i:c:a" ) == Failure ) {
	  /* Give Usage */
	  printf( CYCLE_USAGE );
	  go = 0;
	  break;
	}

    switch( opt ) {
    case 'c': /* Class */
      if( CStoInt32( &optvCS, &val ) == Failure ) {
	  printf( INV_CLASS ); /* Error */
	  go = 0;
      } else class = val;
      break;
      
    case 'i': /* ID */
      if( CStoInt32( &optvCS, &val ) == Failure ) {
	  printf( INV_SERVER );
	  go = 0;
      } else id = val;
      break;
      
    case 's': /* Server */
      if( CSLen( optvCS ) > sizeof(server) - 1 ) 
	CSLen( optvCS ) = sizeof(server) - 1;
      
      strncpy( server, CSStart(optvCS), CSLen(optvCS) );
      server[ CSLen(optvCS) ] = '\0';
      break;

    case 'a': /* Everything */
      break;

    default: 
      printf( CYCLE_USAGE );
      go = 0;
      break;
    }
  }

  if( id >= 0 && class == -1 && server[0] == '\0' ) {
    printf( INV_CLASS );
	go = 0;
  }

  if( go ) cycle_server( class, id );
  
  return Success;
}

/* ------------------------------------------------------------------------- */

static RetValue
cycle_server( long class, long instance )
{
  int     sct    = 0;
  int     fct    = 0;
  ServerParams *sp = NULL;
  LIST   *hlist;

  BootParams bp;

  printf( "\nCycling server processes...\n\n" );
  
  xl_List_init( &hlist, sizeof(ServerParams) );

  /* Compile a list of servers to cycle from BB in one hit...
   */
  compile_cycle_list( class, instance, hlist );
  
  if( xl_List_get_first( hlist, (uchar_t**)&sp ) == Success ) {
    do {
      memcpy( (void*)&(bp.sp), (void*)sp, sizeof(ServerParams) );
      strcpy( bp.domain, local_dom );

	  if( kill_server( &bp ) == Success ) {
        (launch_server( &bp ) == Success) ? sct++ : fct++;
	  } else fct++;
      
    } while( xl_List_get_next( hlist, (uchar_t**)&sp ) == Success );
  } /* endwhile */

  xl_List_end( &hlist );

  printf( "\nCycled %d servers, %d failed\n", sct, fct );
  printf( "\n" );

  return Success;
}

/* ------------------------------------------------------------------------- */

static RetValue
compile_cycle_list( long classid, long srvid, LIST *hlist )
{
  RetValue ret  = Success;
  uchar   *ptr  = NULL;
  SHMLIST *BB   = (SHMLIST*)((char*)bbc + bbc->BB_entry_offset );
  
  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire BB Mutex" );
    return Failure;
  }

  /*
   * Search BB for this server
   */
  if( shm_List_get_first( BB, &ptr ) == Success ) {
    do {
      if( ((BBEntry*)(ptr))->domainid > bbc->domainid ) break;
      
      if( ((BBEntry*)(ptr))->domainid == bbc->domainid ) {

	/* SKIP OVER ADMINISTRATIVE CLASS!!!!! - DON'T WANT TO KILL GUARDIAN!!
	 */
	if( ((BBEntry*)(ptr))->bb_sp.servclass == 1 ) continue;
	
	/* If we want a particular class, and we've overshot, then finish */
	if( classid && ((BBEntry*)(ptr))->bb_sp.servclass > classid ) break;

	/* If we want any class, or a particular class... */
	if( !classid || ((BBEntry*)(ptr))->bb_sp.servclass == classid ) {      
	  /* If we want any server instance, or a particular one... */
	  if( !srvid || srvid == ((BBEntry*)(ptr))->bb_sp.servid ) {

		/* Put the servers ServerParans into our cycle_list
		 */
	    if( xl_List_insert_node( hlist, ListEnd, 
				     (uchar_t*)&(((BBEntry*)(ptr))->bb_sp) ) == Failure ) {
	      ret = Failure;
	      break;
	    }
	  }
	}
      }
    } while( shm_List_get_next( BB, &ptr ) == Success );
  }
  
  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release BB Mutex" );
    return Failure;    
  }
  
  return ret;
}


/* ------------------------------------------------------------------------- */
/* end */
