#ifndef DBCOM_H
#define DBCOM_H
/*
 * Domain Bridge Message Header
 *
 * Copyright 2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * --------------------------------------------------------------------------
 * $Id: dbcom.h,v 1.2 2003/03/29 12:21:10 csmith Exp $
 * --------------------------------------------------------------------------
 */

#define DBM_MAGIC_NUMBER 0x6d447747

/* Define Goldwater Domain Message Version strings
 */
typedef enum { DBM_VERSION_001 = 1 } GDBM_Version;

/* Inter-Domain Message Header offsets
 * All offsets are relative to the start of the message header, so new
 * sections can be added to the header whilst maintaining backwards compat.
 */
/* Standard Control */
#define DBM_MAGIC_NUM_IX   0      /* 4   */
#define DBM_VERSION_IX     4      /* 4   */
#define DBM_MSG_TYPE_IX    8      /* 4   */
#define DBM_FROM_IX        12     /* 32  DOM_NM_MAX */
#define DBM_TO_IX          44     /* 32  DOM_NM_MAX */
#define DBM_SERVICE_IX     76     /* 32  SVC_NM_MAX */
#define DBM_FLAGS_IX       108    /* 4   */
#define DBM_MESSAGE_ID_IX  112    /* 16 (4 words, 128 bits) */
#define DBM_MSG_REP_ID_IX  128    /* 4   */
#define DBM_PRIORITY_IX    132    /* 4   */
#define DBM_TIMEOUT_IX     132    /* 4   */
#define DBM_TXID_IX        140    /* 16 (4 words, 128 bits) */

/* Message Data Section */
#define DBM_DATA_LEN_IX    156    /* 4   */
#define DBM_DATA_OFFSET_IX 160    /* 4   */

/* Authorisation Section */
#define DBM_AUTH_TYPE_IX   164    /* 4   */
#define DBM_AUTH_LEN_IX    168    /* 4   */
#define DBM_AUTH_OFFSET_IX 172    /* 4   */

/* Start of payload */
#define DBM_PAYLOAD_IX     176

typedef enum { 
  DBM_AUTH_NAME = 0x01, 
  DBM_AUTH_MD5  = 0x02, 
  DBM_AUTH_X509 = 0x03
} GDMB_AuthMode;


/* Response Header Offsets */

#define DBR_STATUS_IX      0      /* 4   GWStatus            */
#define DBR_SYS_ERROR_IX   4      /* 4   GWE_MsgFlags & 0xff */
#define DBR_PAYLOAD_IX     8  


# endif

