/*
 * Domain Bridge
 *
 * Copyright 2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * --------------------------------------------------------------------------
 * $Id: dbridge.c,v 1.3 2003/09/13 13:04:32 csmith Exp $
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_linklst.h>
#include <phlib/xl_malloc.h>
#include <phlib/xl_socket.h>

#define GW_INTERNAL_FLAGS 1

#include <gwclient.h>
#include <gwapi.h>
#include <gwlog.h>
#include <gwservice.h>
#include <regstrings.h>
#include <gwreadconf.h>
#include <gwdomain.h>
#include <gwgbprm.h>

#include <gwxlapi.h>
#include <gwneng.h>
#include <bbquery.h>

#include <sessions.h>

#include "version.h"
#include "dbridge.h"
#include "dbctrl.h"

#define TEST_PORT 2441 

const char SERVICE[]  = "GWDOMBRG";

#define STARTUP_MESSAGE1 "GWDBridge Version " DOM_VERSION " Ready."
#define STARTUP_MESSAGE2 "(C)2003 netFluid Technology Ltd"

DomainNode *i_dom_array                     = NULL;   /* Global */
int         i_session_2_domain_map[MAX_FDS] = {-1};

static int data_pending = 0;

static RetValue add_remote_domains();
static RetValue import_remote_services( const char *, int );

static void dump_message_header( GWMsgCtrl *mc );

/* ------------------------------------------------------------------------ */

GWStatus 
GWSRV_INIT( char *s_name, char *reserved )
{
  GWStatus    ret   =  GWSuccess;
  long        queue = -1;
  BBReqQueue *bbq   = NULL;
  BBEntry    *bbe_p = NULL;
  int    ix         = 0;
  BBDom *domarray   = NULL;

  SHMLIST *BB       = (SHMLIST*)((char*)bbc + bbc->BB_entry_offset );

  gw_log_init( _ServId, "GWDBridge" );

  /* Remove and remote domains we think we know about (this is just in case
   * we get restarted.....)
   */
  domarray = (BBDom*)((char *)bbc + bbc->BB_dom_offset);

  /* Start from the first REMOTE domain (ie don't nuke ourselves!!!)
   */
  for( ix = 1; ix < bbc->dommax; ix++ ) {
    *(domarray[ix].domain) = '\0';
	domarray[ix].available = 0;
  }

  if( gw_add_service( "!/dom/services/public", "domsvc1" ) < 0 ) {
    gw_logf( LOG_SERIOUS, "Cannot add service entry someservice" );
    return GWFailure;
  }

  /* Add remote domains and the services imported
   */
  if( add_remote_domains() == Failure ) ret = GWFailure;

  if( init_db_network() == Failure ) ret = GWFailure;

  /* Search for our server entry so that we can find out what our request
   * queue is.  This may seem like a round-a-bout way of getting the msgqid
   * but it avoids breaking the architecture for this one special case.
   */
  if( find_server_entry( bbc->domainid, _ServClass, _ServId, &bbe_p, BB) < 0){
    gw_logf( LOG_SERIOUS, "Could not find Domain Bridge in BB (%ld:%ld:%ld)",
             bbc->domainid, _ServClass, _ServId );
    return GWFailure;
  }

  /* Get reference to BBReqQueue array */
  bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);

  queue = bbq[ bbe_p->req_queue_idx ].msgqid;

  bbc->dbridge_msgq_idx = bbe_p->req_queue_idx;

  gw_logf( LOG_SYSINFO, STARTUP_MESSAGE1 );
  gw_logf( LOG_SYSINFO, STARTUP_MESSAGE2 );

#if 0
  connect_domain( 2 );
  connect_domain( 3 );
#endif

  return ret;
}

/* ------------------------------------------------------------------------ */

void
GWSRV_EXIT()
{
  int    ix       = 0;
  BBDom *domarray = NULL;

  /* Remove services registered against remote domains */

  /* Remove remote domains  (start from the first REMOTE domain!)
   */
  domarray = (BBDom*)((char *)bbc + bbc->BB_dom_offset);

  /* Don't nuke ourselves (dom 0) */
  for( ix = 1; ix < bbc->dommax; ix++ ) {
    *(domarray[ix].domain) = '\0';
	domarray[ix].available = 0;
  }

  gwne_netcom_deinit();
}

/* ------------------------------------------------------------------------ */

static RetValue
add_remote_domains()
{
  RetValue ret    = Success;
  char  path[256] = {0};
  char *domname   = NULL;
  long  len       = 0;
  int   ix        = 0;
  int   r         = 0;
  long maxdom     = 0;
  BBDom *dom      = NULL;

  if( i_dom_array != NULL ) { xl_free( i_dom_array ); i_dom_array = NULL; }

  /* Find out the maximum number of domains */
  if( gw_config_read_long( R_MAX_DOMAINS, &maxdom ) < 0 ) {
    gw_logf( LOG_SERIOUS, "maxdomains not set\n" );
    gwerror = GWE_BADCFG;
    return Failure;
  }
  maxdom++;  /* Let us index from 1..n instead of 0..n-1 */

  if( (i_dom_array = xl_malloc( maxdom * sizeof(DomainNode) )) == NULL ) {
    gw_logf( LOG_SERIOUS, "Cannot allocate memory\n" );
    gwerror = GWE_UNSTABLE;
    return Failure;
  }

  do {
    sprintf( path, "%s%06d%s", R_REMOTE_DOM_B, ix, R_REMOTE_DOM_L );
    gw_logf( LOG_INTDBG, "Installing domain %s", path );

    r = gw_config_read_char( path, &domname, &len );

    if( r >= 0 ) {
      if( (dom = __gw_domain_from_config( domname )) == NULL ) {
	gw_logf( LOG_SYSERR, "Configuration file incomplete for doamin %s",
		 domname );
        ret = Failure;
	/* Carry on with the other remote domains so we can check them all */
      } else {

        if( dom->domainid >= maxdom ) {
          gw_logf( LOG_SERIOUS, "Got a domain ID > maxdom !!! \n" );
          gwerror = GWE_UNSTABLE;
          return Failure;
	}

	/* Add service to import to service list */
        import_remote_services( domname, dom->domainid );

        /* Add domain to internal domain list */
	    i_dom_array[ dom->domainid ].domainid      = dom->domainid;
        i_dom_array[ dom->domainid ].session_id    = -1;
        i_dom_array[ dom->domainid ].conkey        = -1;
        i_dom_array[ dom->domainid ].message_chain = NULL;
        i_dom_array[ dom->domainid ].dead_message_chain = NULL;

	xl_List_init( &i_dom_array[ dom->domainid ].message_chain, sizeof( DomainMessage ) );
	xl_List_init( &i_dom_array[ dom->domainid ].dead_message_chain, sizeof( DomainMessage ) );

      }
    }

    ix ++;

  } while( r >= 0 );

  return ret;
}

/* ------------------------------------------------------------------------ */

static RetValue
import_remote_services( const char *domname, int domainid )
{
  int   id        = 0;
  int   idx       = 0;
  char  path[128] = {0};
  char *service   = NULL;
  long  len       = 0;
  int   ret       = 0;
  
  if( (id = __gw_find_domain_idx_in_config( domname ) ) < 0 ) {
    return Failure;
  }

  idx = 0;
  do {
    sprintf( path, "%s%06d%s%06d%s", 
	     R_DOMAIN_B, id, R_IMPORT_SB, idx, R_IMPORT_L );
    ret = gw_config_read_char( path, &service, &len );

    if( ret == 0 ) {
      if( __gw_add_service_dom( service, "", domainid ) < 0 ) {
        gw_logf( LOG_SERIOUS, "Cannot add service entry %s", service );
        return Failure;
      }
    }
    idx++;
  } while( ret == 0 );

  return Success;
}

/* ------------------------------------------------------------------------ */

int
main( int argc, char **argv )
{
  int   done = 0;
  char *f    = 0;
  char *s    = 0;
  char *d    = 0;
  long fl    = 0;
  long ret   = 0;
  DomainMessage *dm = NULL;
  GWMsgCtrl *rmsg_ctrl = NULL;

  __gw_server_init( argc, argv );

  while( 1 ) {

    data_pending = xlapi_total_writes_pending() | xlapi_events_pending();

    /* Decide whether we need to block or not.  
     * We do a timed block (not a complete) or a non-block.
     */
    fl = data_pending ? GWP_REQNBLK : GWP_REQTBLK;

    ret = __gw_get_request( &f, &s, &d, &fl );

    /* Get pointer to request Ctrl block
     */
    rmsg_ctrl = __gw_get_requestCtrl();

    if( ret >= 0 && !(rmsg_ctrl->cflags & GWC_NOREQM) ) {

      gw_log_init( _ServId, "GWDBridge" );

      if( gw_log_at_level( LOG_INTDBG ) ) {
        /* Dump message header */
        dump_message_header( rmsg_ctrl );
      }

      switch( rmsg_ctrl->msgtype ) {
      case GWM_SVC_REQUEST:
	done = ( __gw_invoke_C_service() < 0 ) ? 0 : 1;
	__gw_complete_request( done ); /* done == 1 will update BB counters*/
	break;
    
      case GWM_DOM_REQUEST:
      case GWM_DOM_REPLY:
	gw_logf( LOG_INTDBG, "inject %s message( &rmsg_ctrl )",
		 rmsg_ctrl->msgtype == GWM_DOM_REQUEST ? "REQUEST" : "REPLY" );

	/* Mark request as coming from a DOMAIN
	 */
	rmsg_ctrl->flags  |= GWP_FROMDOM;
	rmsg_ctrl->cflags |= GWC_FORCERM;  /* Force spoolfile removal when done */

	/* Stop server API moaning that we don't return a reply before ending 
	 * the current request processing (this only applies to domain Req/Reply)
	 */
	rmsg_ctrl->cflags |= GWC_FORWARD;

    /* Connect to domain (just returns if we're already connected)
	 * If we cannot connect to the domain, then we throw away the message
	 */
    if( connect_domain( rmsg_ctrl->target_dom ) == Failure ) {
      gw_logf( LOG_SYSWARN, "Failed to connect to domain %d", 
	       rmsg_ctrl->target_dom );

	  __gw_complete_request( 0 ); /* 0 == Do not update counters */
	  break;
    }

	gw_logf( LOG_INTDBG, "Insert DomainMessage (for domain %d)",
		 rmsg_ctrl->target_dom);
	xl_List_insert_space( i_dom_array[rmsg_ctrl->target_dom].message_chain, ListEnd );

	dm = (DomainMessage*)xl_List_get_current_data( i_dom_array[rmsg_ctrl->target_dom].message_chain );

        /* 
         * We need to 'MOVE' the contents of rmsg_ctrl into dm
         * we do this with a memcpy and then clear out the contets of the
         * original with memset.
         * This is to prevent any cleanup action on rmsg_ctrl freeing any
         * cstrings within rmsg_ctrl etc from messing up the 'shallow copy'
         * we made.
         */
	memcpy( (char*)&dm->rctrl, (char*)rmsg_ctrl, sizeof( GWMsgCtrl ) );
	memset( (char*)rmsg_ctrl, 0, sizeof( GWMsgCtrl ) );
        rmsg_ctrl = NULL; /* Make sure we can't use this any more */

        /* WE CANNOT USE RMSG_CTRL FROM THIS POINT */
        
	dm->q_time = time(NULL);
	dm->ret_error_on_cleanup = GWE_NOERROR;

	gw_logf( LOG_INTDBG, "Become Session" );
	xlapi_become_session_id( i_dom_array[dm->rctrl.target_dom].session_id);

	gw_logf( LOG_INTDBG, "Register Event for session %d",
		 i_dom_array[ dm->rctrl.target_dom].session_id );
	xlapi_register_event( EV_USER_EVENT, EU_NEW_MESSAGE);

	break;
    
      default:
	gw_logf( LOG_INTDBG, "Unexpected message type %ld",
		 rmsg_ctrl->msgtype );
      }
    
    }

    gw_log_init( _ServId, "GWDBridge" ); 
    gwne_netcom_poll( 1, 0 ); /* 1 loop, don't block */
  }

  exit(0);
}

/* ------------------------------------------------------------------------ */

void
domsvc1( char *svc, GWMessage *reqm )
{
  gw_logf( LOG_INFO, "Domain Test Service '%s' Invoked", svc); 
   
  return;
}

/* ------------------------------------------------------------------------ */

static void
dump_message_header( GWMsgCtrl *mc )
{
  char * str = "";
  char flags_s[256] = {0};
  char uuidbuf[64];

  gw_logf( LOG_INTDBG, "mctrl.total_size    = %ld", mc->total_size );
  gw_logf( LOG_INTDBG, "mctrl.bsize         = %ld", mc->bsize );
  gw_logf( LOG_INTDBG, "mctrl.ident         = %ld", mc->ident );
  gw_logf( LOG_INTDBG, "mctrl.destsrvid     = %ld", mc->destsrvid );
  gw_logf( LOG_INTDBG, "mctrl.target_dom    = %ld", mc->target_dom );

  switch( mc->msgtype ) {
  case GWM_SVC_REQUEST : { str = "GWM_SVC_REQUEST"; break; }
  case GWM_SVC_REPLY   : { str = "GWM_SVC_REPLY";   break; }
  case GWM_DOM_REQUEST : { str = "GWM_DOM_REQUEST"; break; }
  case GWM_DOM_REPLY   : { str = "GWM_DOM_REPLY";   break; }
  case GWHM_HALT_SERVER: { str = "GWHM_HALT_SERVER"; break; }
  default: { str = "Unknown Message Type"; break; }
  }

  uuid_unparse( mc->msgid, uuidbuf );
  gw_logf( LOG_INTDBG, "mctrl.msgtype       = %s", str );
  gw_logf( LOG_INTDBG, "mctrl.host          = %s", mc->host ? mc->host:"None" );
  gw_logf( LOG_INTDBG, "mctrl.port          = %ld", mc->port );
  gw_logf( LOG_INTDBG, "mctrl.msgqid        = %ld", mc->msgqid );
  gw_logf( LOG_INTDBG, "mctrl.r_msgqid      = %ld", mc->r_msgqid );
  gw_logf( LOG_INTDBG, "mctrl.org_r_msgqid  = %ld", mc->org_r_msgqid );
  gw_logf( LOG_INTDBG, "mctrl.msgid         = %s",  uuidbuf );
  gw_logf( LOG_INTDBG, "mctrl.clid          = %ld", mc->clid );
  gw_logf( LOG_INTDBG, "mctrl.spool_dir     = %s",  mc->spool_dir );
  gw_logf( LOG_INTDBG, "mctrl.service       = %s",  mc->service );
  gw_logf( LOG_INTDBG, "mctrl.function      = %s",  mc->function);


  flags_s[0] = '\0';
  if( mc->flags & GW_NOREPLY )  strcat( flags_s, "GW_NOREPLY | " );
  if( mc->flags & GW_NOBLOCK )  strcat( flags_s, "GW_NOBLOCK | " );
  if( mc->flags & GWP_ASYNC )   strcat( flags_s, "GWP_ASYNC | " );
  if( mc->flags & GWP_DFILE )   strcat( flags_s, "GWP_DFILE | " );
  if( mc->flags & GWP_REPLY )   strcat( flags_s, "GWP_REPLY | " );
  if( mc->flags & GWP_PHXRET )  strcat( flags_s, "GWP_PHXRET | " );
  if( mc->flags & GWP_LRGMSG )  strcat( flags_s, "GWP_LRGMSG | " );
  if( mc->flags & GWP_REMOTE )  strcat( flags_s, "GWP_REMOTE | " );
  if( mc->flags & GWP_TIAVM )   strcat( flags_s, "GWP_TIAVM | " );
  if( mc->flags & GWP_HOMDMSG ) strcat( flags_s, "GWP_HOMDMSG | " );
  if( mc->flags & GWP_REQNBLK ) strcat( flags_s, "GWP_REQNBLK | " );

  gw_logf( LOG_INTDBG, "mctrl.flags         = %s",  flags_s   );

  flags_s[0] = '\0';
  if( mc->cflags & GWC_SPOOL )   strcat( flags_s, "GWC_SPOOL | " );
  if( mc->cflags & GWC_PORTAL )  strcat( flags_s, "GWC_PORTAL | " );
  if( mc->cflags & GWC_DONE )    strcat( flags_s, "GWC_DONE | " );
  if( mc->cflags & GWC_REPSENT ) strcat( flags_s, "GWC_REPSENT | " );
  if( mc->cflags & GWC_RDONLY)   strcat( flags_s, "GWC_RDONLY | " );
  if( mc->cflags & GWC_NOBLOCK ) strcat( flags_s, "GWC_NOBLOCK | " );
  if( mc->cflags & GWC_NOREQM )  strcat( flags_s, "GWC_NOREQM | " );
  if( mc->cflags & GWC_FORWARD ) strcat( flags_s, "GWC_FORWARD | " );

  gw_logf( LOG_INTDBG, "mctrl.cflags        = %s",  flags_s   );

  gw_logf( LOG_INTDBG, "Data Length         = %ld", CSLen(mc->cache) );
  gw_logf( LOG_INTDBG, "Spool Filename      = %s",  mc->spoolfile );
  gw_logf( LOG_INTDBG, "+ Bytes Spooled     = %ld", mc->bytes_spooled );

}

/* ------------------------------------------------------------------------ */
/* Warning.  This is complicated - at first glance.
 */
RetValue
connect_domain( int domid )
{
  const char *sid        = NULL;
  int         fd         = 0;
  int         session_id = 0;
  ServiceID   svcid      = -1;
  BBDom *domarray     = NULL;

  domarray = (BBDom*)((char *)bbc + bbc->BB_dom_offset);
      
  /* First check if we've got a connection to the domain already...
   */
  /* Take 'session_id' from i_dom_array[domid] and check the service_id
   * associated with this session.
   */
  session_id = i_dom_array[domid].session_id;

  /* If we've got a +ve session ID, see if 'netcom' thinks that this is
   * associated with a service or not.
   * If it is, and it's connected, then it will return the service string
   *   if it's not connected, then it will return NULL.
   * Just a check that the string matches the service that we're expecting
   * it to be connected to is enough to see if we have a connection.
   *
   * The connection may have dropped between calls to gwne_netcom_poll()
   * and we'll only detect this once we try a write/read on the fd, so
   * this check is just whether we *think* we've got a connection.
   *
   * (We've only got one service being handled by gwne_netcom* so the
   *  SERVICE check is purely academic - but it is "The Right Thing To Do").
   */
  if( session_id >= 0 ) sid = xlapi_get_service_id( session_id );

  if( sid == NULL || strcmp( sid, SERVICE ) != 0 ) {
    /* This session is not connected, or is connected to the wrong service. */
    session_id = -1;
    i_dom_array[domid].session_id = -1;
  }

  /* Okay, we've established that the fd stored against domid IS connected
   * to a service of type SERVICE.  What we haven't done is verified that
   * it is connected to the expected DOMAIN!
   *
   * (ie two remote domains A and B are connected on fd's 4 and 5.
   *  Both domains drop.  The fd's are still associated with domains A and B.
   *  We send a message to domain B, which gets reconnected but with fd 4.
   *  We then send a message to domain A.  The fd against this domain is the
   *  stale '4'.  When we do the above validation, we discover that fd 4 is
   *  connected to a remote domain, and being handled by service SERVICE.
   *  But we THINK it's connected to domain A, when in fact it's connected to
   *  domain B ! )
   *
   * The fix is to maintain a mapping between fd and Domain.
   * We've got a mapping from domain -> fd, and the reverse map will verify
   * that the connection really is valid.
   */

  if( session_id >= 0 ) {
    if( i_session_2_domain_map[session_id] != domid ) {

	  gw_logf( LOG_INTDBG, "Session not connected to the expected domain - reconnecting" );

	  /* Ooops, this session is not connected to the domain we wanted.
	   * Force a reconnect and a new session to handle it.
	   */
      session_id = -1;
      i_dom_array[domid].session_id = -1;
	}
  }

  if( session_id == (-1) ) {

    /* Get Service ID of Domain Bridge Network Handling */
    if( (svcid = xlapi_get_serviceID( SERVICE )) < 0 ) {
      gw_logf( LOG_SERIOUS, "Cannot find handlers for netservice %s", SERVICE );
      return Failure;
    }

    gw_logf( LOG_INTDBG, "Getting connection to domain %s [%s:%d]", domarray[domid-1].domain, domarray[domid-1].gw_ipaddr, domarray[domid-1].gw_port );

    /* we've not got a connection to the remote domain.  Get one. 
     */
    fd = xl_get_connection( domarray[domid-1].gw_ipaddr, domarray[domid-1].gw_port );
	
    if( fd < 0 ) {
      gw_logf( LOG_SYSWARN, "Cannot establish connection with domain %s [%s:%d]", domarray[domid-1].domain, domarray[domid-1].gw_ipaddr, domarray[domid-1].gw_port );
      return Failure;
    }

    gw_logf( LOG_SYSINFO, "Connection established with domain %s [%s:%d]", domarray[domid-1].domain, domarray[domid-1].gw_ipaddr, domarray[domid-1].gw_port );
	 
    session_id = fd;  /* Session ID's are fds */

    i_dom_array[domid].session_id      = session_id;
    i_session_2_domain_map[session_id] = domid;

    domarray[ domid - 1 ].available = 1; /* Mark as available */

    /* When a new connection is registered event EV_NEW_SESSION is invoked
     * with whatever 'value' given in  arg[1] below passed as *private.
     * This can be an integer or even a ptr to memory.
     * A new session usually allocates a chunk of private WS memory and passes
     * uses *private to pass it back and forth.
     *
     * We need to get 'domid' through to the new session we're creating.
     * So there are two choices:
     * 1) Allocate a DomSession WS here and populate it with domid and other
     *    values and pass &DomSession through xlapi_register_event()
     * 2) Pass domid through xlapi_register_event() directly.
     *
     * The later is prefered as we only have one value to pass through to
     * the new session, and it means that all session memory allocation/freeing
     * is done within the session handler code itself.  This lets us keep
     * details of what is contained within a sessions 'private' WS memory
     * private!
     * 
     */

    /* Register Connection to netEngine will handle it for us.
     * Pass 'domid' through as private data so the new session can index back
     * to the data it needs in the BB and the i_domain_array.
     */
    xlapi_register_connection( fd, (void*)domid, svcid );
  }

  return Success;

}

/* ------------------------------------------------------------------------ */
/* end */
