/*
 * Goldwater Distributed Web Architecture.
 *
 * Billboard Query Routines
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:10 $
 * 
 * --------------------------------------------------------------------------
 * $Log: bbquery.c,v $
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.25  2003/01/04 19:34:11  chris
 * gw_init() is now quite intelligent with respect to reconnecting.
 *
 * Revision 1.24  2003/01/04 16:28:48  chris
 * gw_init is now intelligent about detecting if Goldwater is running.
 * By repeatedly calling gw_init() a client can make sure they stay attached
 * without having to do gw_init(), gw_release() over and over.  gw_init() will
 * detect that the backend has been bounced and try its hardest to reconnect.
 *
 * Revision 1.23  2002/09/01 10:59:31  chris
 * Addition of namespaceing to service name.
 * Service may now be of the form this.is.a.namespace/and/this/is/the/service
 * Wildcards are NOT allowed in RPC type message destinations.
 *
 * Revision 1.22  2002/08/04 19:36:03  chris
 * query_bb now allows scanning of a single domain's services.
 * API changes to allow replies to be sent back from dombridge.
 *
 * Revision 1.21  2002/08/01 20:49:25  chris
 * Domain IDS are from 1..n, domain array index is 0..n-1 -> made sure that all
 * domainid indexes into array are 0..n-1
 *
 * Revision 1.20  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.19  2002/06/04 16:35:45  chris
 * Introduce target_dom
 *
 * Revision 1.18  2002/05/26 15:32:20  chris
 * Stop breaking out when service class does not match - services are
 * ordered by domainid, so the same class may re-occur.
 *
 * Revision 1.17  2002/05/14 21:40:56  chris
 * Removed blank lines
 *
 * Revision 1.16  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.15  2002/05/12 13:16:22  chris
 * Cleanup of unnecessary fields in BBEntry.
 * (Created new structure ServerParams that contains the fields common
 *  to those stored in the BB and those in the BootParam message).
 *
 * Revision 1.14  2002/03/17 18:44:07  chris
 * Handle the destruction of shm by exposing shmid value.
 *
 * Revision 1.13  2002/03/11 22:25:48  chris
 * Bugfixes: Getting in a twist when collecting stale reply packets and
 * corrupting the ctrl structure.  Fixed stale packet loop logic and made
 * sure ctrl struct side effects are documented and considered.
 *
 * Revision 1.12  2002/02/24 18:05:38  chris
 * Moved APPDIR to BBControl block instead of duplicating it for each and
 * every server.
 * Return GWE_APPERR and set gwusrerr when Services return failure.
 * Use new regstrings that describe the new XML based configuration.
 * Added new functions to support individual addition of service entries.
 * New Version 1.14
 *
 * Revision 1.11  2002/01/03 20:27:31  chris
 * Introduction of preliminary Domain Control handling.
 *
 * Revision 1.10  2001/12/26 22:57:42  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.9  2001/06/03 20:04:39  chris
 * Problems with attach_bb - didn't seem to detect error return of -1
 * correctly.  This might be due to linking with the wrong version of
 * gwlib, but sorted out the casting around shmat so that there is no
 * doubt that it'll get it right!
 *
 * Revision 1.8  2001/06/03 10:48:16  chris
 * attach_bb now non-static.  shmat returns -1 in an error, not NULL.
 * Fixed this.
 *
 * Revision 1.7  2001/04/17 13:36:36  chris
 * All find functions now consider and sort by domainid.
 * New function find_domain (moved from billboard.c in guardian).
 *
 * Revision 1.6  2001/04/15 12:54:16  chris
 * Introduced domain table.
 *
 * Revision 1.5  2001/03/30 17:22:17  chris
 * srch_for_class_id was returning -1 if a match wasn't made on class and
 * id.  This is not what was intended as this is a 'where should we
 * insert?' search function!  Now returns (properly) where a class/id
 * server was expected to be found in the list.
 *
 * Revision 1.4  2001/03/26 14:37:45  chris
 * Pass SRV_LIST into srch_for_class_id.  find_server_entry now sets
 * *s_ent to point to the last entry that matched on class, regardless of
 * whether it gets an ID match.
 *
 * Revision 1.3  2001/03/26 14:18:56  chris
 * Added srch_for_class_id and optimised find_server_entry with the
 * knowledge that server entries are ordered on class and id.
 *
 * Revision 1.2  2001/03/26 10:32:42  chris
 * Shm permissions changed to 0600
 *
 * Revision 1.1  2001/03/19 15:58:59  chris
 * Initial revision
 *
 * --------------------------------------------------------------------------
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#include <phlib/xl_types.h>

#include <goldreq.h>
#include <mutex.h>
#include <gwlog.h>
#include <bbquery.h>

static BBControl *_bbc = NULL;
extern int gwerror;

/* Prototypes
 */
BBControl * get_bb_control( key_t shmipc );

/* ------------------------------------------------------------------------- */

/* Find a service in the service table.
 * If we're being called from a server to find another service, than DO NOT
 * match against ourselves!!!
 *
 * The least selected service from multiple instances is selected.
 * (Simple round-robbin).
 *
 * Real clients should supply -1 for rqsrvcls and rqsrvid's (as they won't
 * have values for these parameters).
 */

RetValue
query_bb( const char *service, BBSvc **sinfo, BBDom **dinfo, 
	  SHMLIST *SVC_List, long rqsrvcls, long rqsrvid, long wantdomid )
{
  RetValue ret     = Failure;
  uchar   *ptr     = NULL;
  BBSvc   *mptr    = NULL;

  unsigned long minused = -1;
  
  gwerror = GWE_NOSVC;

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire SVC Mutex" );
    gwerror = GWE_SYSFAIL;
    return Failure;    
  }
  
  minused = 0xffffffff;

  if( _bbc == NULL ) {
    gw_logf( LOG_SERIOUS, "Not attached to BB when querying" );
    gwerror = GWE_SYSFAIL;
    return Failure;
  }

  /* Try and find the requested service in the service map.
   * NOTE: There may be more than one entry, and may be on different
   *       machines.  Thus basic Load Balancing where the least picked service
   *       entry is chosen.
   */
  if( shm_List_get_first( SVC_List, &ptr ) == Success ) {
    do {
      if( strcmp( ((BBSvc *)ptr)->service, service) == 0 ) {
	
	if( (((BBSvc*)ptr)->servclass == rqsrvcls &&
	    ((BBSvc*)ptr)->servid    == rqsrvid ) || 
		(wantdomid && ((BBSvc*)ptr)->domainid != wantdomid) ) {
	  /* 
	   * This is a service _within_ our own server... we can't
	   * call this one...  OR we want services for a specific domain
	   * ... Look for another.
	   */
	  gwerror = GWE_SBLOCKD; /* Service is blocked from access */
	  continue; /* Find next match */
	}

	/* Keep hold of the service match that we've picked the least
	 */
	if( ((BBSvc*)ptr)->picked < minused  ) {
	  mptr    = (BBSvc*)ptr; 
	  minused = mptr->picked; 

	  ret = Success;
	}
      }
    } while( shm_List_get_next( SVC_List, &ptr ) == Success ) ;
  }

  if( ret == Success ) {
    /* Found matching service name */
    mptr->picked++;
    *sinfo = mptr;

    /* Domain info associated with this service
     */
    *dinfo = &((BBDom*)((char *)_bbc + _bbc->BB_dom_offset))[ mptr->domainid - 1];

    gwerror = 0;
  }
  
  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release SVC Mutex" );
    gwerror = GWE_SYSFAIL;
    return Failure;    
  }

  return ret;
}

/* ------------------------------------------------------------------------- */

SHMLIST *
get_service_list( key_t shmipc )
{
  /* Attach to SHM
   */
  if( _bbc == NULL && get_bb_control( shmipc ) == NULL ) return NULL;
  
  return (SHMLIST*)((char*)_bbc + _bbc->BB_svc_offset );  
}

/* ------------------------------------------------------------------------- */

SHMLIST *
get_server_list( key_t shmipc )
{
  /* Attach to SHM
   */
  if( _bbc == NULL && get_bb_control( shmipc ) == NULL ) return NULL;
  
  return (SHMLIST*)((char*)_bbc + _bbc->BB_entry_offset );  
}

/* ------------------------------------------------------------------------- */

BBControl *
get_bb_control( key_t shmipc )
{
  int shmid = 0;
  char *res = NULL;

  if( _bbc == NULL ) {
    if( (_bbc = attach_bb( shmipc, &shmid )) == NULL ) {
      res =  "Cannot attach to SHM segment.";
	  write( 2, res, strlen(res) );
      return NULL;
    }
  }

  return _bbc;
}

/* ------------------------------------------------------------------------- */
/* Drop attachmet to SHM
 */
void
release_bb_control()
{
  if( _bbc != NULL ) shmdt( _bbc );
  _bbc = NULL;
}

/* ------------------------------------------------------------------------- */
/*
 * Tries to attach to a shm segment.
 * If it doesn't exist, or shmat fails then NULL is returned
 */
BBControl *
attach_bb( key_t ipckey, int *shmid )
{
  long bbmem = -1;
  
  /* SHM MUST already Exist!
   */
  *shmid = shmget( ipckey, 0, 0600 );

  /* shmat (really annoyingly!) returns -1 on failure, else a ptr
   * to the attached segment..... so checking for NULL failure return
   * fails. Argg!
   */
  if( *shmid < 0 ||
      (bbmem = (long)shmat( *shmid, 0, 0 )) < 0 ) return NULL;

  return (BBControl *)bbmem;
}


/* ------------------------------------------------------------------------- */
/* Find a given service entry within a particular Server Class within a
 * particular domain.
 * If server ID is specified, then the service entry for the given class and
 * id is located.
 */
int
find_services_for_server( long domainid, long class, long id, 
			  const char *service,
			  BBSvc **sinfo, SHMLIST *SVC_List )
{
  int    ret = -1;
  uchar *ptr = NULL;

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire SVC Mutex" );
    return -1;    
  }

  /* Find the specified service entry within the collection of service ads 
   * for this server class......
   * List is ordered by domain, class and id, so we can abort the search
   * whenever domain, class or id goes too big.
   */
  if( shm_List_get_first( SVC_List, &ptr ) == Success ) {
    do {
      if( ((BBSvc *)ptr)->domainid > domainid ) break;
      if( ((BBSvc *)ptr)->domainid < domainid ) continue;

      /* Got domain... */
      if( ((BBSvc *)ptr)->servclass > class ) continue;
      if( ((BBSvc *)ptr)->servclass < class ) continue;

      /* Got Domain and class... */
      if( id == (-1) || ((BBSvc*)ptr)->servid == id ) {
	if( strcmp( ((BBSvc *)ptr)->service, service) == 0 ) {
	  ret = 1;
	  break;
	}
      }
    } while( shm_List_get_next( SVC_List, &ptr ) == Success );
  }

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release SVC Mutex" );
    return -1;    
  }

  if( ret ) *sinfo = (BBSvc*)ptr;

  return ret;
}

/* ------------------------------------------------------------------------- */
/*
 * Searches BB for a server with a given class and id.
 * Either an exact match is returned, or the position in the list where
 * that server should be inserted.
 */
int
srch_for_class_id( long domainid, long class, long id, SHMLIST *SRV_List  )
{
  long     pos   = -1;
  BBEntry *s_ent = NULL;

  /*
   * TODO: Hash table!
   */

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ){
    gw_logf( LOG_SERIOUS, "Cannot aquire mutex" );
    return -1;
  }

  /*
   * Search BB for this server
   */
  p_find_server_entry( domainid, class, id, &s_ent, &pos, SRV_List );

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Cannot release mutex" );
    return -1;
  }

  return pos; 
}

/* ------------------------------------------------------------------------- */
/*
 * Searches for a server of domain, class and ID within the shm.
 * If it finds a class match it sets s_ent to the last matching class it
 * found.
 * If it also gets an ID match, then TRUE is returned, else -1 is returned.
 */
int
find_server_entry( long domainid, long class, long id,
		   BBEntry **s_ent, SHMLIST *SRV_List )
{
  int    ret = -1;
  long   pos = 0;

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire BB Mutex" );
    return -1;    
  }
  
  /* Find the specified service entry within the collection of service ads 
   * for this server class......
   */
  ret = p_find_server_entry( domainid, class, id, s_ent, &pos, SRV_List );

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release BB Mutex" );
    return -1;    
  }

  return ret;
}

/* ------------------------------------------------------------------------- */
/*
 * Searches for a server of domain, class and ID within the shm.
 * If it finds a class match it sets s_ent to the last matching class it
 * found.
 * pos is returned set to the position in SRV_List that the exact match was
 * found or the position in the list where the server should be inserted.
 * pos is ListEnd if list was empty.
 *
 * 1 is returned for exact matches, -1 otherwise.
 */
int
p_find_server_entry( long domainid, long class, long id,
		     BBEntry **s_ent, long *pos, SHMLIST *SRV_List )
{
  int      ret   = -1;
  uchar   *ptr   = NULL;

  *s_ent = NULL;

  /*
   * Search BB for this server
   */
  *pos = ListEnd;

  if( shm_List_get_first( SRV_List, &ptr ) == Success ) {
    *pos = 0;
    do {
      if( ((BBEntry*)(ptr))->domainid > domainid ) break;
      
      if( ((BBEntry*)(ptr))->domainid == domainid ) {
	if( ((BBEntry*)(ptr))->bb_sp.servclass > class ) break;
	
	if( ((BBEntry*)(ptr))->bb_sp.servclass == class ) {
	  
	  *s_ent = (BBEntry*)ptr;
	  
	  if( ((BBEntry*)(ptr))->bb_sp.servid >= id ) {
	    /* Found a server with at least the same id
	     */
	    if( ((BBEntry*)(ptr))->bb_sp.servid == id ) ret = 1; /* EXACT! */
	    break;
	  }
	}
      }
      (*pos)++;
    } while( shm_List_get_next( SRV_List, &ptr ) == Success ) ;
  }
  
  return ret;
}

/* ------------------------------------------------------------------------- */

BBDom *
__gw_find_domain( const char *domain, BBControl *bbc )
{
  BBDom *domarray = NULL;
  int    ix       = 0;
  
  domarray = (BBDom*)((char *)bbc + bbc->BB_dom_offset);
  
  for( ix = 0; ix < bbc->dommax; ix++ ) {
    if( *(domarray[ix].domain) != '\0' ) {
      if( strcmp( domarray[ix].domain, domain ) == 0 ) {
	    /* Found domain */
	    return &(domarray[ix]);
      }
    }
  }
  
  /* Not found */
  return NULL;
}

/* ------------------------------------------------------------------------- */

BBDom *
__gw_find_domain_by_namespace( const char *namespace, BBControl *bbc )
{
  BBDom *domarray = NULL;
  int    ix       = 0;

  domarray = (BBDom*)((char *)bbc + bbc->BB_dom_offset);

  for( ix = 0; ix < bbc->dommax; ix++ ) {
    if( *(domarray[ix].domain) != '\0' ) {
      if( strcmp( domarray[ix].namespace, namespace ) == 0 ) {
	    /* Found domain */
	    return &(domarray[ix]);
      }
    }
  }
  
  /* Not found */
  return NULL;
}

/* ------------------------------------------------------------------------- */
/* end */

