/*
 * Goldwater Distributed Web Architecture.
 *
 * User Client Command API
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.5 $  $Date: 2003/09/13 13:04:32 $
 * 
 * --------------------------------------------------------------------------
 * $Log: gwclient.c,v $
 * Revision 1.5  2003/09/13 13:04:32  csmith
 * Portable.net PInvoke helper functions.
 * Guardian bug: Did not clean BBReqQueue table on failure/shutdown of servers.
 * Increased version numbers to reflect significant bug fixes.
 * Announce copyright FSF in administrator.
 * Administrator now pauses 1/10 second between boot/halt commands to give
 * guardian time to breath (not strictly necessary, but may avoid some signal
 * race conditions).
 *
 * Revision 1.4  2003/06/12 21:39:54  csmith
 * Preliminary ASYNC request work.  Pnet C# wrappers.
 *
 * Revision 1.3  2003/05/29 22:39:17  csmith
 * Added some TODO's for async handling
 *
 * Revision 1.2  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.42  2003/03/24 22:12:32  chris
 * Allow forcing of disconnect (ignore reference count).
 *
 * Revision 1.41  2003/01/04 19:34:11  chris
 * gw_init() is now quite intelligent with respect to reconnecting.
 *
 * Revision 1.40  2003/01/04 16:28:48  chris
 * gw_init is now intelligent about detecting if Goldwater is running.
 * By repeatedly calling gw_init() a client can make sure they stay attached
 * without having to do gw_init(), gw_release() over and over.  gw_init() will
 * detect that the backend has been bounced and try its hardest to reconnect.
 *
 * Revision 1.39  2003/01/04 12:36:55  chris
 * made sure umask is <0777
 *
 * Revision 1.38  2003/01/03 00:09:11  chris
 * Use internal umask when creating files, shm and queues
 *
 * Revision 1.37  2002/12/28 19:07:10  chris
 * Call to __gw_init was not using return values.
 * Added check that guardian is running to __gw_init
 *
 * Revision 1.36  2002/11/11 20:26:34  chris
 * implemented gw_svc_forward.
 *
 * Revision 1.35  2002/11/10 12:23:41  chris
 * Cleanup and low level API changes for bindings such as C#
 *
 * Revision 1.34  2002/10/20 13:38:34  chris
 * Enabled setting of loglevel from administrator (removed _LogLevel and
 * replaced it with *_LogLevelPtr through out code).
 *
 * Revision 1.33  2002/10/20 11:57:42  chris
 * Namespace extraction from service name does not work if the service call
 * is a homed message (the service name is a data structure in this case).
 *
 * Revision 1.32  2002/09/01 11:26:09  chris
 * Removed some debug
 *
 * Revision 1.31  2002/09/01 10:59:31  chris
 * Addition of namespaceing to service name.
 * Service may now be of the form this.is.a.namespace/and/this/is/the/service
 * Wildcards are NOT allowed in RPC type message destinations.
 *
 * Revision 1.30  2002/08/27 18:52:15  chris
 * Removed obsolete code
 *
 * Revision 1.29  2002/08/26 00:29:54  chris
 * More control for domain handling (like forcing the removal of spoolfiles)
 *
 * Revision 1.28  2002/08/04 19:36:03  chris
 * query_bb now allows scanning of a single domain's services.
 * API changes to allow replies to be sent back from dombridge.
 *
 * Revision 1.27  2002/06/16 11:40:35  chris
 * BB Client is now the Domain Bridge message tracking list.
 * New error codes and messages.
 *
 * Revision 1.26  2002/06/16 00:18:31  chris
 * Use gwuuid_compare to compare uuid_t buffers (4 32bit ints) because it is
 * faster.
 *
 * Revision 1.25  2002/06/15 19:32:05  chris
 * Message ID values are now 128 bit OSF DCE compatible uuid's.
 *
 * Revision 1.24  2002/06/11 21:52:58  chris
 * Enabled timed blocking in gwp_get_request and use this in dbridge to
 * prevent spinning and chewing cpu time.
 *
 * Revision 1.23  2002/06/10 18:06:18  chris
 * Make sure SIGALARM is unblocked.
 *
 * Revision 1.22  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.21  2002/06/09 20:19:26  chris
 * More TODO and comment changes
 *
 * Revision 1.20  2002/06/04 16:35:45  chris
 * Introduce target_dom
 *
 * Revision 1.19  2002/06/02 19:26:42  chris
 * Maintain count of pending requests (requests in queue).
 *
 * Revision 1.18  2002/05/30 19:37:34  chris
 * Create entries in BBClient table for each new client attaching to
 * application.
 *
 * Revision 1.17  2002/05/26 15:36:46  chris
 * If we're a SHM mode client, and the service being called is not local
 * then mark the message as type GWM_DOM_REQUEST and dispatch it (it will
 * go to the domain bridge queue because that queue will allways be set
 * against the remote services in the BB).
 *
 * Revision 1.16  2002/05/12 17:46:16  chris
 * Store msgid of expected reply in SHM so that Domain Bridge can validate
 * the reply at it's door ..
 *
 * Revision 1.15  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.14  2002/04/21 20:21:41  chris
 * TEMPORARY: Use GWNET environment variable to force clients into
 * NET mode.
 *
 * Revision 1.13  2002/04/14 00:02:26  chris
 * *** empty log message ***
 *
 * Revision 1.12  2002/04/12 19:33:10  chris
 * Change to msgsnd setup so that non-homed messages are sent with an ident
 * of 1.  This is because a homed message and a service request placed in a
 * queue together, both with min(listen_ident) ident values *may* result in
 * the service request being passed over.
 *
 * Revision 1.11  2002/03/17 11:19:00  chris
 * Allow request lengths of ZERO
 *
 * Revision 1.10  2002/03/12 22:51:03  chris
 * Added BBReqQueue support.  Currently defines a BBReqQueue entry for
 * each server, even though servers are grouped into classes and thus
 * will usually share a queue.
 *
 * Revision 1.9  2002/03/11 22:25:48  chris
 * Bugfixes: Getting in a twist when collecting stale reply packets and
 * corrupting the ctrl structure.  Fixed stale packet loop logic and made
 * sure ctrl struct side effects are documented and considered.
 *
 * Revision 1.8  2002/03/10 19:35:33  chris
 * Handles proper homed messages (initially for shutdown of servers).
 *
 * Revision 1.7  2002/03/02 19:09:39  chris
 * Handle skipping of stale messages in the queue.
 *
 * Revision 1.6  2002/02/24 18:05:38  chris
 * Moved APPDIR to BBControl block instead of duplicating it for each and
 * every server.
 * Return GWE_APPERR and set gwusrerr when Services return failure.
 * Use new regstrings that describe the new XML based configuration.
 * Added new functions to support individual addition of service entries.
 * New Version 1.14
 *
 * Revision 1.5  2001/12/26 22:56:51  chris
 * New style of goldwter client config file (uses [SECTION] sections)
 *
 * Revision 1.4  2001/04/15 12:56:42  chris
 * Introduced domain table.
 *
 * Revision 1.3  2001/03/26 18:49:05  chris
 * Removed all chain_ident code, replacing most with destsrvid code.
 *
 * Revision 1.2  2001/03/22 01:18:56  chris
 * Added GWC_RDONLY flag to GWMessage flags.  This prevents GWMsg structs
 * which are wrapped around a request message from having the memory the
 * request is sitting in messed about with, if the GWMsg is used as a
 * service call reply holder.  It will seriously screw up the
 * request_cleanup code if it is messed with!
 *
 * Revision 1.1  2001/03/19 16:04:16  chris
 * Initial revision
 *
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <phlib/xl_memlst.h>
#include <phlib/xl_malloc.h>
#include <phlib/xl_mmap.h>

#define GW_INTERNAL_FLAGS 1

#include <gwmsg.h>
#include <goldreq.h>

#include <gwreadconf.h>
#include <gwportal.h>
#include <bbquery.h>

#define GW_PRV_CLIENT 1
#include <gwclient.h>
#undef  GW_PRV_CLIENT

#include <gwtime.h>
#include <gwgbprm.h>
#include <gwlog.h>
#include <gwdomain.h>
#include <gwuuid.h>
#include <regstrings.h>

static GWMessage cs_gwm = {0};
static uuid_t instanceID;

#ifdef GWNETWORK
static long static_log_level = 1;
#endif

/* Prototypes
 */
static int get_local_domain_idx();

/* ------------------------------------------------------------------------- */
/*
 * Initialise CLIENT.  This establishes a connection with Goldwater and hooks
 * up with the BB (if available).
 *
 * Environment Variables Used:
 * GWCONFIG = Goldwater config file supplying:
 *            - mode of operation shm or net
 *            - IPCKEY if shm
 *            - Host/Port if net
 *            - Logfile directory ( just if net ? )
 */
__inline__ int
__gw_init( )
{
  int   ret       = -1;
  char *res       = NULL;
  char  path[256] = {0};

  static int dom_idx   = -1;

  if( _ServId >= 0 ) return 0; /* Don't do anything if we're a server */

  if( _gw_initialised < 0 ) _gw_initialised = 0; /* Saftey net */

  if( gw_config_init() < 0 ) return -1;

  _Mode = (char*)MODE_SHM;  /* legacy stuff */

  do {

    /* Only get dom_idx once. 
     */
    if( dom_idx < 0 && (dom_idx = get_local_domain_idx()) < 0 ) break;
  
    /*
     * get IPCKey once only ( _IpcKey is global, -1 initially)
     */
    if( _IpcKey < 0 ) { 
      sprintf( path, "/goldwater/domain/%06d/ipckey", dom_idx );
      if( gw_config_read_long( path, &_IpcKey ) < 0 ) {
        gwerror = GWE_BADIPCKY;
        break;
      }
      if( _IpcKey > MAXIPCKEY ) {
        gwerror = GWE_BADIPCKY;
        break;
      } 
    }
  
  
    if( _gw_initialised ) {
      if( bbc && gwuuid_compare( instanceID, bbc->instance_id ) == 0 ) {
  
        /* We ARE already initialised - 
	 * increment our reference count and return !
	 */
        _gw_initialised++;
  
        ret = 0;
	break;
      } 
  
      /* Okay, we're basically doing a re-connect here
       * Drop any attached resources
       */
      res = "Goldwater was stopped whilst I was connected. Attempting a re-connect.\n";
      gw_logf( LOG_SYSWARN, res ); /* Log it if logging is configured */
      write( 2, res, strlen(res) );

      gwp_remove_reply_queue();
      gwp_portal_disconnect();
      detach_bbc_mutex();   /* To force a re-attach */
      release_bb_control();
      __gw_reset_global_params();
      bbc = NULL;
    }

    /* IpcKey is still valid if we ever get this far (cached).
     */
    if( attach_bbc_mutex( _IpcKey ) == Failure ) {
      res =  "Goldwater is not running.  Cannot continue.\n";
      gw_logf( LOG_SYSWARN, res ); /* Log it if logging is configured */
      gwerror = GWE_NOTUP;
      break;
    }
  
    /* Now see if guardian is running.  If not, DO NOT go near SHM! 
     */
    if( get_bbc_mutex_value( MUTEX_GUARDIAN ) > 0 ) {
      res =  "Goldwater is not running.  Cannot continue.\n";
      gw_logf( LOG_SYSWARN, res ); /* Log it if logging is configured */
      gwerror = GWE_NOTUP;
      break;
    }

    if( __gw_umask_from_config( dom_idx ) == Failure ) break;
    
    /* Now handle SHM connection and lookup of static parameters, such as
     * spool directory, log directory, maxmsgtx etc
     *
     * This also increments _gw_initialised reference counter
     */
    if( __gw_set_global_params() < 0 ) break;
  
    /* Pre-init log */
    gw_log_init( 9999, "client" );
    gw_logf( LOG_AUDIT, "Client joining Goldwater application" );
  
    /* Record the attached instance ID 
     */
    uuid_copy( instanceID, bbc->instance_id );

    ret = 0;

  } while ( 0 );

  gw_config_deinit();

  return ret;
}


/* ------------------------------------------------------------------------- */

static int 
get_local_domain_idx()
{
  int dom_idx = -1;
  long len = 0;
  char *res       = NULL;

  /* Find out what our local domain is...
   */
  if( gw_config_read_char( "/goldwater/control/local/local", &res, &len ) < 0 ){
    gwerror = GWE_BADCNFG;
    return -1;
  }

  /* Discover what branch the local domain is in... R_DOMAIN_B
   */
  if( (dom_idx = __gw_find_domain_idx_in_config(res) ) < 0 ) {
    gwerror = GWE_BADCNFG;
  }

  return dom_idx;
}

/* ------------------------------------------------------------------------- */

int
gw_init( )
{
  int ret = 0;

  if( _ServId >= 0 ) return 0; /* Don't do anything if we're a server */

  ret = __gw_init();

  /*
   * Register function to remove reply queue on exit
   * (This may not work when API is wrapped (eg Perl/C#)
   *  so these bindings need to call gw_release themselves).
   */
  if( _gw_initialised == 1 ) atexit( (void(*)())gw_release );

  return ret;
}

/* ------------------------------------------------------------------------- */
/*
 * Detach Goldwater client
 * - Ignore the reference count if FORCE is true (necessary where we may be
 *   forced to do multiple gw_init() calls, but only one gw_release()
 */
void
gw_release( int force )
{ 
  if( _ServId < 0 && _gw_initialised > 0  ) {

    _gw_initialised--;

    /* Only detatch from Goldwater instance when reference count hits zero */

    if( force || _gw_initialised == 0 ) {
      gw_logf( LOG_AUDIT, "Client leaving Goldwater application" );

      gwp_remove_reply_queue();
      gwp_portal_disconnect();
    
      _gw_initialised = 0;
    }
  }
}

/* ------------------------------------------------------------------------- */
/*
 * Send a request to a goldwater service and get a reply back.
 *
 *
 */
__inline__ int
gw_svc_call( const char *service, const char *reqdata, long reqlen,
	     GWMessage *replymsg, long flags )
{
  GWMsgCtrl   ctrl  = {0};
  RetValue    retv  = Success;
  BBSvc      *sinfo = NULL;
  BBDom      *dinfo = NULL;
  BBReqQueue *bbq   = NULL;
  const char *svcname = NULL;
  char        namespace[ NAME_SPACE_MAX + 1 ] = {0};
  sigset_t    smask;

  long       bsent  = 0;
  long       left   = 0;
  long       tosend = 0;
  long       bsize  = 0;
  long       wantdom= 0;
  int        ix     = 0;

  uuid_t     msgid;

  static   pid_t mypid  = 0;
  unsigned long  keep_flags = 0;


  if( _gw_initialised == 0 ) {
    gw_logf( LOG_SYSERR, "Not connected to Goldwater" );
    gwerror = GWE_NOTCNCT;
    return -1;
  }

  if( reqlen < 0 ) {
    gw_logf( LOG_SYSERR, "Bad request length" );
    gwerror = GWE_BADDATA;
    return -1;
  }

  /* Create unique message id */
  uuid_generate( msgid );

  gwerror = 0;

  /* Service may contain a namespace reference.  Extract and lookup if
   * necessary.
   *
   * my.namespace.declaration/some/service/name
   */
  wantdom = 0;
  ix = 0;
  /* Only scan service if we know it is a real service name and not a homed
   * data block
   */
  while( !(flags & GWP_HOMDMSG ) && service[ix] > 0 && service[ix] != '/' && service[ix] != '!') ix++;
  svcname = service + ix;
  if( ix ) {
    if( ix > NAME_SPACE_MAX ) {
      gw_logf( LOG_SYSERR,"Namespace exceeds maximum length allowed (%d) [.*s]",
	       NAME_SPACE_MAX, NAME_SPACE_MAX, service  );
      gwerror = GWE_BADNSPC;
      return -1;
    }
    if( ix ) strncpy( namespace, service, ix );
    namespace[ix] = '\0';
    dinfo = __gw_find_domain_by_namespace( namespace, bbc );
    if( dinfo == NULL ) {
      gw_logf( LOG_SYSERR, "Unknown namespace [%s]", namespace );
      gwerror = GWE_BADNSPC;
      return -1;
    }
    wantdom = dinfo->domainid;
  }


  /* Keep only the flags that are valid for this api function.
   */
  ctrl.flags = flags & ( GW_NOREPLY | GW_NOBLOCK | GWP_ASYNC );

  ctrl.msgqid      = -1;

  ctrl.total_size  = reqlen;
  ctrl.bsize       =  _MaxMsgTx;

  ctrl.ident       = -1;  /* Filled in if we're in a SHM model */
  ctrl.destsrvid   = -1; 
  ctrl.target_dom  = 1; 

  ctrl.clid        = mypid;     /* Client ID    */
  uuid_copy( ctrl.msgid, msgid );
  ctrl.host        = _GWHost;   /* Local Domain netServer Hostname/IP */
  ctrl.port        = _GWPort;   /* Local Domain netServer Port        */
  ctrl.spool_dir   = _SpoolDir;
  ctrl.service     = (char*)svcname;

  ctrl.msgtype = GWM_SVC_REQUEST;

  WrapCStringLen( ctrl.cache, "", 0 );

  if( strcmp( _Mode, MODE_SHM ) == 0 ) {
    /* Lookup service in SHM.
     *
     * It *may* come back as being remote if it exists on a remote machine.
     *
     */
    if( flags & GWP_HOMDMSG ) {
      /* Function has been called by a GOLDWATER INTERNAL system specifying
       * that the request is targetted at a SPECIFIC server instance.
       * In which case (char*)service points to a HomedMsg structure which
       * referenced the target server.
       *
       * Chris: I admit this is a little hacky, but is only used in a handful
       *        of instances and is probably safe because normal users don't
       *        get to see the GWP_* flags in the header files.
       */
      ctrl.msgqid    = ((GWHomedMsg*)service)->msgqid;
      ctrl.destsrvid = ((GWHomedMsg*)service)->servid;
      ctrl.ident     = ((GWHomedMsg*)service)->ident; 
      ctrl.msgtype   = ((GWHomedMsg*)service)->msgtype;
      ctrl.function  = "control";
      ctrl.service   = "control";
      
      ctrl.flags |= GWP_HOMDMSG;
      ctrl.flags |= GWP_TIAVM;
      
    } else {
      /* Called by a user application.  Do the lookup in the BB.
       */
      if( query_bb( svcname, &sinfo, &dinfo,
		    _ServiceList, _ServClass, _ServId, wantdom ) == Failure) {
        gw_logf( LOG_SYSWARN,
	         "Service not found or blocked (gwerror %ld)", gwerror );
        return -1;
      } 

      /* Get reference to BBReqQueue array */
      bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);

      /* Set MIN message Ident that all servers are looking for and get the
       * msgqid to send the request to...
       */
      
      /* TODO: Different values may be used here to provide some sort of 
       * priority to messages
       */
      ctrl.ident    = 1; /* Lowest possible! */

      ctrl.msgqid   = bbq[ sinfo->req_queue_idx ].msgqid;
      
      ctrl.function   = sinfo->function;
      ctrl.target_dom = sinfo->domainid; /* Essential for dombridge routing */
      
      /* Increment count of requests on this queue
       */
      __gw_increment_queue_count( sinfo->req_queue_idx );
      
      if( sinfo->flags & GWS_REMOTE_SERVICE ) {
        /* Service has been found and is available on a remote goldwater
         * Domain.  The request for this queue will be the request queue
	 * of the Domain Bridge.  We need to indicate the type of message
	 * being sent to the Domain Bridge so that it knows the direction
	 * of travel (request through bridge, or reply through bridge).
         */
	ctrl.msgtype = GWM_DOM_REQUEST;
      }
    }
  } else {
    /* else mode is net...
     *
     * This may be deprecated with the new Domain handling currently
     * being introduced 01/05/2002.......
     */
    ctrl.flags |= GWP_REMOTE;
  }


  /* If we get passed a message block for reply that has been used before
   * then we need to either release or re-use the memory buffer within
   * to avoid a leak!  We just release the memory if we're a mapped spoolfile.
   */
  if( !(flags & GW_NOREPLY) && replymsg->data != NULL ) {
    if( replymsg->_pflg & GWC_SPOOL ) {
      /* We've got a mapped spoolfile... just release it */
      gw_free_message( replymsg );
    }
    else {
      /* Wrap a cstring around a malloced memory area, allowing it to use
       * it as though IT had malloced it... Unless that is, it's read only!!!
       */
      if( replymsg->_pflg & GWC_RDONLY ) {
	/* GWMsg contains a RDOnly data area.  
	 * Point to some re-usable memory instead.
	 */
	WrapCStringLen( ctrl.cache, "", 0 );

	gw_logf( LOG_INTDBG, "ReqMsg contains RDONLY data.  Creating new." );
      } else {
	/* Not read only... we're safe to re-use the memory.
	 */
	MirrorCString( ctrl.cache, replymsg->_bcx );
	gw_logf( LOG_INTDBG, "ReqMsg contains Reusable data... Using it!" );
      }
    }
  }

  if( gwp_write_block_init( &ctrl ) == Failure ) {
    /* gwerror will be set to a suitable error code */
    return -1;
  }

  /* We're a message originator, so set our originator Reply Queue
   * (This is used exclusively by GWDBridge/netServer to get replies back to
   *  the intended client).
   */
  ctrl.org_r_msgqid = ctrl.r_msgqid;

  /* TODO: block 'chunking' of data is no longer required as msgqueue messages
   *       are not 'chained'.
   *       This 'bsize' stuff and the loop can probably be removed.
   */

  /* If we're sending the reply to the NetServer 
   * OR we're spooling to disk, then do it all as one chunk.
   */
  bsize = ctrl.bsize;
  if( (ctrl.cflags & GWC_PORTAL) || (ctrl.cflags & GWC_SPOOL)) bsize = reqlen;

  /*
   * Initiate timeout
   */
  sigemptyset( &smask );
  sigaddset  ( &smask, SIGALRM );
  sigprocmask( SIG_UNBLOCK, &smask, NULL );

  signal( SIGALRM, __got_timeout );

  switch( sigsetjmp( timeout_env, 1 ) ) {
    
  case TIMER_START:
    alarm( _ReqTout );
    gw_logf( LOG_INTDBG, "gwclient.c: Setting service call timeout to %ld", 
	     _ReqTout );
    gw_logf( LOG_INTDBG, "gwclient.c: Sending request to service: %s",
	     ctrl.service);
    
    bsent = 0;
    do { 
      left   = reqlen - bsent;
      tosend = left >= bsize ? bsize : left;
      
      if( gwp_write_block_data( &ctrl, reqdata + bsent, tosend ) == Failure ) {
	gw_logf(LOG_SYSERR, "Failed to send request to Goldwater service %s",
		ctrl.service );

	ctrl.flags &= ~GWP_REPLY;
	gwp_cleanup_ctrl( &ctrl );

	signal( SIGALRM, SIG_IGN );
	alarm(0);
	return -1;
      }
      bsent += tosend;
    } while( bsent < reqlen );
    
    retv = gwp_write_block_flush( &ctrl );

    signal( SIGALRM, SIG_IGN );
    alarm(0);      

    if( retv == Failure ) {
      ctrl.flags &= ~GWP_REPLY; /* We want the spoolfile removed.. */
      gwp_cleanup_ctrl( &ctrl );
      return -1;
    }
    break;

  case TIMER_EXPIRED: 
    gw_logf( LOG_SYSERR, 
	     "Timeout occured when sending request to service: %s", ctrl.service );
    
    ctrl.flags &= ~GWP_REPLY;
    gwp_cleanup_ctrl( &ctrl );

    /* Decrement count of requests on this queue as it never got anywhere
     */
    __gw_decrement_queue_count( sinfo->req_queue_idx );
    
    gwerror = GWE_TIMEOUT;
    signal( SIGALRM, SIG_IGN );
    return -1;
    break;
  }


  /* If the request specified no reply or is ASync, then just return 
   */
  if( flags & (GW_NOREPLY | GWP_ASYNC) ) return 0;

  /* Now get reply.  The reply will be sent with it's msgid set to the msgid 
   * that we sent with the request.  This msgid is unique, and thus 
   * guarantees that we can always find our reply in a reply queue that has
   * 'stale' replies in it.
   */

  ctrl.flags &= ~GWE_SVCUFAIL;
  keep_flags  = ctrl.flags;

  signal( SIGALRM, __got_timeout );
  switch( sigsetjmp( timeout_env, 1 ) ) {
    
  case TIMER_START:
    alarm( _ReqTout );    
    gw_logf( LOG_INTDBG, "gwclient.c: Get reply from service: %s", ctrl.service );
    
    do {
      gw_logf( LOG_INTDBG, "Wait for reply....." );
      
      /* These get trampled on by gw_get_reply, so always set before a call
       */
      uuid_copy( ctrl.msgid, msgid ); /*Make sure gwp_get_reply gets OUR reply*/
      ctrl.flags = keep_flags;
      
      if( gwp_get_reply( &ctrl ) == Failure ) {
	gw_logf( LOG_SYSERR, "gwp_get_reply failed" );
	signal( SIGALRM, SIG_IGN );
	alarm(0);
	return -1;
      }
    } while( gwuuid_compare( ctrl.msgid, msgid ) != 0 ); /* Wait for reply to CURRENT message*/
    
    signal( SIGALRM, SIG_IGN );
    alarm(0);
    break;

  case TIMER_EXPIRED:
    gw_logf( LOG_SYSERR,
	     "Timeout occured whilst waiting for reply from service: %s",
	     ctrl.service );
    gwerror = GWE_TIMEOUT;
    return -1;
    break;
  }

  /*
   * Return GWMessage block containing all the info we need to manage the
   * allocated resources....
   */
  MirrorCString( replymsg->_bcx, ctrl.cache );

  replymsg->_pflg = 0;          /* Safe to clear RDONLY flag now */
  replymsg->data  = CSStart( ctrl.cache );
  replymsg->len   = CSLen( ctrl.cache );

  if( ctrl.cflags & GWC_SPOOL ) {
    replymsg->_pflg |= GWC_SPOOL;
    strcpy( replymsg->_ctx, ctrl.spoolfile );
  }

  if( ctrl.flags & GWE_SVCUFAIL ) {
    /* Service returned an error
     */

    if( ctrl.flags & GWE_DOMFAIL ) {
      /* Internal domain error - return in gwerror */
      gwerror = ctrl.flags & GWE_ERRMASK; /* Mask off the error flags */
    } else {
      /* General Application Level Failure */
      gwusrerr = ctrl.flags & GWE_ERRMASK; /* Mask off the error flags */
      gwerror  = GWE_APPERR;
    }

    return -1;
  }

  return 0;
}

#if ENABLE_ASYNC
/* ------------------------------------------------------------------------- */
/* Perform an asynchronous request.
 */
int
gw_svc_call_async( const char *service, const char *reqdata, long reqlen,
		   GWAsyncHandle *msgid, long flags )
{
  /* TODO call gw_svc_call with GWP_ASYNC and capture the msgid
   * storing it in GWAsyncHandle
   */
}
#endif

/* ------------------------------------------------------------------------- */
/* Special wrapper of gw_svc_call for those callers that don't want to know
 * about GWMessages.
 * 
 * The memory used in the reply is not freed until
 * __gw_cs_svc_call_cleanup() is called - however, gw_svc_call will try and
 * re-use the memory if it is not freed before then. Thus it's probably okay
 * to defer the cleanup of cs_gwm until you're all done.
 *
 * SPOOLFILES will remain on disk until gw_cs_svc_call_cleanup is called, or
 * another call to __gw_cs_svc_call is made.
 *
 * Repeat use of this function therefore overwrites the data returned in 
 * previous invokations.
 */
int
__gw_cs_svc_call( const char *service, const char *reqdata, long reqlen,
                  char **replydata, long *replylen, long flags )
{
  if( gw_svc_call( service, reqdata, reqlen, &cs_gwm, flags ) < 0 ) {
    return -1;
  }

  *replydata = cs_gwm.data;
  *replylen  = cs_gwm.len;

  return 0;
}

/* ------------------------------------------------------------------------- */

void
__gw_cs_svc_call_cleanup()
{
  gw_free_message( &cs_gwm );
}

/* ------------------------------------------------------------------------- */

void
gw_free_message( GWMessage *gwm  )
{
  if( gwm->_pflg & GWC_SPOOL ) {
    
    gwm->_pflg &= ~GWC_SPOOL;      /* Blit flag      */
    munmap( gwm->data, gwm->len ); /* Unmap the file */
    unlink( gwm->_ctx );           /* Remove file    */
    gwm->data = NULL;              /* Clear Pointer  */
  }
  
  /* If we've got some *real* data, and the message is not RDONLY
   * then we can free it.  If it is RDONLY, then it belongs to someone
   * else, and they'll be freeing it (like the request handler).
   */
  if( gwm->data != NULL && !(gwm->_pflg & GWC_RDONLY) ) {
    xl_free( gwm->data );
  }
  gwm->data = NULL;                /* Clear Pointer  */
  gwm->len  = 0;                   /* Reset length   */
  
  WrapCString( gwm->_bcx, "" );    /* Clear Cstring  */
}

#if 0

/* ------------------------------------------------------------------------- */
/*
 * Get a client ID by searching the BBClient list for a free slot
 * (List items are ordered by client ID, so as soon as there is a gap in the
 *  sequence, we have a free slot).
 */
static BBClient*
get_client_id()
{
  int       ix      = 0;
  long      last_id = 0;
  uchar    *ptr     = NULL;
  BBClient *client  = NULL;
  
  SHMLIST *ClientList = (SHMLIST*)((char*)bbc + bbc->BB_clt_offset );

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC_CLT ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire CLT Mutex" );
    return NULL;
  }

  ix      = 0;
  last_id = 0;

  if( shm_List_get_first( ClientList, &ptr ) == Success ) {
    do {
      if( ((BBClient *)ptr)->id > (last_id + 1) ) break; 
	  
      last_id = ((BBClient *)ptr)->id;

    } while( ++ix && shm_List_get_next( ClientList, &ptr ) == Success ) ;
  }

  do {
    /* ix points to the free slot, unless ix > bbc->clientmax, in which case
     * there is no room in the client list.
     */
    if( ix > bbc->clientmax ) {
      gw_logf( LOG_ERROR, 
	       "Maximum number of clients attached to application (%ld)", 
	       bbc->clientmax );
      break;
    } 

    if( shm_List_insert_space( ClientList, ix ) == Failure ) {
      gw_logf( LOG_SERIOUS, 
	       "Could not add client %ld to Client table", last_id + 1 );
      break;
    }

    /* Get pointer to (empty) node
     */
    client = (BBClient *)shm_List_get_current_data( ClientList );

    /* Claim the slot */
    client->id           = last_id + 1;
    client->class        = getpid();
    client->expect_msgid = -1;
    client->repmqid      = -1;

  } while(0);

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC_CLT ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release CLT Mutex" );
    return NULL;
  }

  return client;
}

/* ------------------------------------------------------------------------- */

static RetValue
release_client_id( BBClient **client ) 
{
  int    ix  = 0;
  uchar *ptr = NULL;

  SHMLIST *ClientList = (SHMLIST*)((char*)bbc + bbc->BB_clt_offset );

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC_CLT ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire CLT Mutex" );
    return Failure;
  }

  ix = 0;
  if( shm_List_get_first( ClientList, &ptr ) == Success ) {
    do {
      if( ((BBClient *)ptr)->id > (*client)->id ) break; 
      if( ((BBClient *)ptr)->id < (*client)->id ) continue; 

      shm_List_delete_node( ClientList );
      *client = NULL;
      break;

    } while( shm_List_get_next( ClientList, &ptr ) == Success ) ;
  }

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC_CLT ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release CLT Mutex" );
    return Failure;
  }

  return Success;
}

/* ------------------------------------------------------------------------- */

static BBClient *
find_client_slot( )
{
  return NULL;
}

#endif

/* ------------------------------------------------------------------------- */

int
__gw_increment_queue_count( long idx )
{
  BBReqQueue *bbq   = NULL;

  bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire BB Mutex" );
    return -1;    
  }

  bbq[ idx ].numonqueue ++;
  
  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release BB Mutex" );
    return -1;    
  }
  return 0;
}

/* ------------------------------------------------------------------------- */

int
__gw_decrement_queue_count( long idx )
{
  BBReqQueue *bbq   = NULL;

  bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);

  /* Get Mutex */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not aquire BB Mutex" );
    return -1;    
  }

  bbq[ idx ].numonqueue --;
  
  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Could not release BB Mutex" );
    return -1;    
  }
  return 0;
}

/* ------------------------------------------------------------------------- */

RetValue
__gw_umask_from_config( int domlev )
{
  char *mask_s = NULL;
  char *ep     = NULL;
  char  path[256] = {0};
  long  mask_l = 0;
  long  len    = 0;

  sprintf( path, "%s%06d%s", R_DOMAIN_B, domlev, R_UMASK_L );

  if( gw_config_read_char( path, &mask_s, &len ) < 0 ) mask_s = "066";

  /* Umask is usually an octal number, so we may need to convert it.
   * We use a ZERO base so that the umask may be given in Decimal
   * (where base 10 is assumed), hex (if 0x prefix is given) or Octal (if
   * a leading 0 is given).
   */
  mask_l = strtol( mask_s, &ep, 0 );
  if( *ep != '\0' || mask_l > 0777 ) {
    gw_logf( LOG_SERIOUS, "Invalid umask in configuration" );
    gwerror = GWE_BADCNFG;
    return Failure;
  }
  _Mask = mask_l & 0777; /* Yeah, okay paranoid about this value! */

  return Success;
}

/* ------------------------------------------------------------------------- */
/* end */
