/*
 * Goldwater Distributed Web Architecture.
 *
 * Global parameters
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:11 $
 * 
 * --------------------------------------------------------------------------
 * $Log: gwgbprm.c,v $
 * Revision 1.2  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.17  2003/01/04 19:34:11  chris
 * gw_init() is now quite intelligent with respect to reconnecting.
 *
 * Revision 1.16  2003/01/04 12:36:55  chris
 * made sure umask is <0777
 *
 * Revision 1.15  2003/01/03 00:09:11  chris
 * Use internal umask when creating files, shm and queues
 *
 * Revision 1.14  2002/11/10 12:23:41  chris
 * Cleanup and low level API changes for bindings such as C#
 *
 * Revision 1.13  2002/10/20 13:38:34  chris
 * Enabled setting of loglevel from administrator (removed _LogLevel and
 * replaced it with *_LogLevelPtr through out code).
 *
 * Revision 1.12  2002/06/16 11:40:35  chris
 * BB Client is now the Domain Bridge message tracking list.
 * New error codes and messages.
 *
 * Revision 1.11  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.10  2002/05/30 19:37:34  chris
 * Create entries in BBClient table for each new client attaching to
 * application.
 *
 * Revision 1.9  2002/05/12 17:46:16  chris
 * Store msgid of expected reply in SHM so that Domain Bridge can validate
 * the reply at it's door ..
 *
 * Revision 1.8  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.7  2002/02/24 18:05:38  chris
 * Moved APPDIR to BBControl block instead of duplicating it for each and
 * every server.
 * Return GWE_APPERR and set gwusrerr when Services return failure.
 * Use new regstrings that describe the new XML based configuration.
 * Added new functions to support individual addition of service entries.
 * New Version 1.14
 *
 * Revision 1.6  2002/01/29 22:43:40  chris
 * Cleaned up warnings.
 * Added gwp_ prefix to shutdown_server.
 * Added regdata.[ch] config entry to build.
 *
 * Revision 1.5  2001/12/26 22:57:43  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.4  2001/06/30 16:41:05  chris
 * Added __gw_populate_globals_from_shm
 *
 * Revision 1.3  2001/04/01 22:45:59  chris
 * Initialise _MaxMsgTX to -1. Moved _MyServEntry into gbprm from
 * gwapi.c.  Removed LrgFlThrsh.
 *
 * Revision 1.2  2001/03/19 17:42:17  chris
 * CFG_* parameters (static const chars) were being included in multiple
 * files.
 *
 * Revision 1.1  2001/03/19 16:05:06  chris
 * Initial revision
 *
 * --------------------------------------------------------------------------
 */
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <phlib/xl_memlst.h>
#include <goldreq.h>
#include <gwportal.h>
#include <bbquery.h>
#include <mutex.h>
#include <gwclient.h>

#define HOSTMAX 31

int   _gw_initialised   = 0;
char *_Mode             = (char*)MODE_SHM;
char *_LogName          = NULL;
char *_SpoolDir         = NULL;
char *_GWHost           = NULL;
char *_Domain           = NULL;

long  *_LogLevelPtr     = NULL;
long  _GWPort           = 0;
long  _IpcKey           = -1;
long  _MaxMsgTx         = -1;
long  _ReqTout          = 30;

long  _ServClass        = -1;
long  _ServId           = -1;

int   _ReplyQueue       = -1;

mode_t _Mask            = 066;

BBEntry   *_MyServEntry = NULL;
/* BBClient  *_MyClientEntry = NULL; */

SHMLIST   *_ServiceList = NULL;
SHMLIST   *_ServerList  = NULL;

BBControl *bbc          = NULL;
 
/* ------------------------------------------------------------------------- */

void
__gw_populate_globals_from_shm()
{
  _SpoolDir    = bbc->spooldir;
  
  _MaxMsgTx    = bbc->maxmsgtx;
  
  _LogName     = bbc->logname;
  _LogLevelPtr = &bbc->loglevel;

  _ReqTout     = bbc->reqtout;
  
  _GWPort      = bbc->gwport;
  _GWHost      = bbc->gwhost;

  _Mask        = bbc->umask & 0777;
}

/* ------------------------------------------------------------------------- */

void
__gw_reset_global_params()
{
  _SpoolDir    = NULL;
  
  _MaxMsgTx    = -1;
  
  _LogName     = NULL;
  _LogLevelPtr = NULL;

  _ReqTout     = -1;
  
  _GWPort      = 0;
  _GWHost      = NULL;

  _Mask        = 066;
}
/* ------------------------------------------------------------------------- */

int
__gw_set_global_params()
{
  /* Now handle SHM connection and lookup of static parameters, such as
   * spool directory, log directory, maxmsgtx etc
   */
  if( strcmp( _Mode, MODE_SHM ) == 0 ) {

    if( (bbc = get_bb_control( _IpcKey )) == NULL ) {
      gwerror = GWE_SYSFAIL;
      return -1;
    }

    __gw_populate_globals_from_shm();

    if( attach_bbc_mutex( _IpcKey ) == Failure ) {
      gwerror = GWE_SYSFAIL;
      return -1;
    }

    if( (_ServiceList = get_service_list( _IpcKey )) == NULL ) {
      gwerror = GWE_SYSFAIL;
      return -1;
    }

    if( (_ServerList = get_server_list( _IpcKey )) == NULL ) {
      gwerror = GWE_SYSFAIL;
      return -1;
    }
  }

  _gw_initialised++;

  return 0;
}

/* ------------------------------------------------------------------------- */
/* end */
