/*
 * Goldwater Domain Network Server
 *
 * Copyright 2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * --------------------------------------------------------------------------
 * $Id: nserver.c,v 1.2 2003/03/29 12:21:11 csmith Exp $
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_linklst.h>
#include <phlib/xl_malloc.h>
#include <phlib/xl_socket.h>

#define GW_INTERNAL_FLAGS 1

#include <gwclient.h>
#include <gwapi.h>
#include <gwlog.h>
#include <gwservice.h>
#include <regstrings.h>
#include <gwreadconf.h>
#include <gwdomain.h>
#include <gwgbprm.h>

#include <gwxlapi.h>

#include <gwneng.h>

#include "version.h"
#include "nsctrl.h"

#define STARTUP_MESSAGE1 "netServer Version " VERSION " Ready."
#define STARTUP_MESSAGE2 "(C)2002 netFluid Technology Ltd"

static void dump_message_header( GWMsgCtrl *mc );

/* ------------------------------------------------------------------------ */

GWStatus 
GWSRV_INIT( char *s_name, char *reserved )
{
  GWStatus ret =  GWSuccess;

  gw_log_init( _ServId, "GWNServer" );

  if( init_ns_network() == Failure ) ret = GWFailure;

  gw_logf( LOG_SYSINFO, STARTUP_MESSAGE1 );
  gw_logf( LOG_SYSINFO, STARTUP_MESSAGE2 );

  return ret;
}

/* ------------------------------------------------------------------------ */

void
GWSRV_EXIT()
{
}

/* ------------------------------------------------------------------------ */

int
main( int argc, char **argv )
{
  int   done = 0;
  char *f    = 0;
  char *s    = 0;
  char *d    = 0;
  long fl    = 0;
  long ret   = 0;
  GWMsgCtrl *rmsg_ctrl = NULL;

  __gw_server_init( argc, argv );

  while( 1 ) {

    /* Most of the action seen by this server will be with fd's - so we'll
	 * choose to block on a select if we have to and not a __gw_get_request.
	 */
    fl = GWP_REQNBLK; 

    ret = __gw_get_request( &f, &s, &d, &fl );


    /* Get pointer to request Ctrl block
	 */
	rmsg_ctrl = __gw_get_requestCtrl();

	if( ret >= 0 && !(rmsg_ctrl->cflags & GWC_NOREQM) ) {

      /* Dump message header */
      dump_message_header( rmsg_ctrl );

      switch( rmsg_ctrl->msgtype ) {
        case GWM_SVC_REQUEST:
          done = ( __gw_invoke_C_service() < 0 ) ? 0 : 1;
          __gw_complete_request( done ); /* done == 1 will update BB counters*/
	    break;
    
		default:
		  gw_logf( LOG_INTDBG, "Unexpected message type %ld",
		           rmsg_ctrl->msgtype );
          __gw_complete_request( 0 ); /* done == 1 will update BB counters*/
        break;
      }
    }

    gw_log_init( _ServId, "GWNServer" );
    gwne_netcom_poll( 1, 2 ); /* 1 loop, 2 = block for 2 seconds */
  }

  exit(0);
}

/* ------------------------------------------------------------------------ */

static void
dump_message_header( GWMsgCtrl *mc )
{
  char * str = "";
  char flags_s[256] = {0};
  char uuidbuf[64];

  gw_logf( LOG_INTDBG, "mctrl.total_size    = %ld", mc->total_size );
  gw_logf( LOG_INTDBG, "mctrl.bsize         = %ld", mc->bsize );
  gw_logf( LOG_INTDBG, "mctrl.ident         = %ld", mc->ident );
  gw_logf( LOG_INTDBG, "mctrl.destsrvid     = %ld", mc->destsrvid );
  gw_logf( LOG_INTDBG, "mctrl.target_dom    = %ld", mc->target_dom );

  switch( mc->msgtype ) {
  case GWM_SVC_REQUEST : { str = "GWM_SVC_REQUEST"; break; }
  case GWM_DOM_REQUEST : { str = "GWM_DOM_REQUEST"; break; }
  case GWM_DOM_REPLY   : { str = "GWM_DOM_REPLY";   break; }
  case GWHM_HALT_SERVER: { str = "GWHM_HALT_SERVER"; break; }
  default: { str = "Unknown Msg Type"; break; }
  }

  uuid_unparse( mc->msgid, uuidbuf );
  gw_logf( LOG_INTDBG, "mctrl.msgtype       = %s", str );
  gw_logf( LOG_INTDBG, "mctrl.host          = %s", mc->host ? mc->host:"None" );
  gw_logf( LOG_INTDBG, "mctrl.port          = %ld", mc->port );
  gw_logf( LOG_INTDBG, "mctrl.msgqid        = %ld", mc->msgqid );
  gw_logf( LOG_INTDBG, "mctrl.r_msgqid      = %ld", mc->r_msgqid );
  gw_logf( LOG_INTDBG, "mctrl.msgid         = %s",  uuidbuf );
  gw_logf( LOG_INTDBG, "mctrl.clid          = %ld", mc->clid );
  gw_logf( LOG_INTDBG, "mctrl.spool_dir     = %s",  mc->spool_dir );
  gw_logf( LOG_INTDBG, "mctrl.service       = %s",  mc->service );
  gw_logf( LOG_INTDBG, "mctrl.function      = %s",  mc->function);


  flags_s[0] = '\0';
  if( mc->flags & GW_NOREPLY )  strcat( flags_s, "GW_NOREPLY | " );
  if( mc->flags & GW_NOBLOCK )  strcat( flags_s, "GW_NOBLOCK | " );
  if( mc->flags & GWP_ASYNC )   strcat( flags_s, "GWP_ASYNC | " );
  if( mc->flags & GWP_DFILE )   strcat( flags_s, "GWP_DFILE | " );
  if( mc->flags & GWP_REPLY )   strcat( flags_s, "GWP_REPLY | " );
  if( mc->flags & GWP_PHXRET )  strcat( flags_s, "GWP_PHXRET | " );
  if( mc->flags & GWP_LRGMSG )  strcat( flags_s, "GWP_LRGMSG | " );
  if( mc->flags & GWP_REMOTE )  strcat( flags_s, "GWP_REMOTE | " );
  if( mc->flags & GWP_TIAVM )   strcat( flags_s, "GWP_TIAVM | " );
  if( mc->flags & GWP_HOMDMSG ) strcat( flags_s, "GWP_HOMDMSG | " );
  if( mc->flags & GWP_REQNBLK ) strcat( flags_s, "GWP_REQNBLK | " );

  gw_logf( LOG_INTDBG, "mctrl.flags         = %s",  flags_s   );

  flags_s[0] = '\0';
  if( mc->cflags & GWC_SPOOL )   strcat( flags_s, "GWC_SPOOL | " );
  if( mc->cflags & GWC_PORTAL )  strcat( flags_s, "GWC_PORTAL | " );
  if( mc->cflags & GWC_DONE )    strcat( flags_s, "GWC_DONE | " );
  if( mc->cflags & GWC_REPSENT ) strcat( flags_s, "GWC_REPSENT | " );
  if( mc->cflags & GWC_RDONLY)   strcat( flags_s, "GWC_RDONLY | " );
  if( mc->cflags & GWC_NOBLOCK ) strcat( flags_s, "GWC_NOBLOCK | " );
  if( mc->cflags & GWC_NOREQM )  strcat( flags_s, "GWC_NOREQM | " );
  if( mc->cflags & GWC_FORWARD ) strcat( flags_s, "GWC_FORWARD | " );

  gw_logf( LOG_INTDBG, "mctrl.cflags        = %s",  flags_s   );

  gw_logf( LOG_INTDBG, "Data Length         = %ld", CSLen(mc->cache) );
  gw_logf( LOG_INTDBG, "Spool Filename      = %s",  mc->spoolfile );
  gw_logf( LOG_INTDBG, "+ Bytes Spooled     = %ld", mc->bytes_spooled );

}

/* ------------------------------------------------------------------------ */
/* end */
